{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE TypeApplications #-}
module AppFiller where

import Import
import Types

import qualified Data.Text as T

import Data.Time
import Data.FingerTree
import Data.HashMap.Strict ((!?))
import qualified Data.HashMap.Strict as HM


import qualified Debug.Trace as D

appFiller :: App -> IO ()
appFiller app = runRIO app $ withRunInIO $ \run -> do
  let queue = twsConnectionRecieve . twsConnectionRefs . appRefs $ app
      debugMsg x = run $ logDebug (display $ T.pack $ "FILLER  : " <> T.unpack x)
      infoMsg  x = run $ logDebug (display $ T.pack $ "FILLER  : " <> T.unpack x)
  forever $ do
    input <- atomically $ readTQueue queue
    let currentAppData = appData app
    case input of
      (Msg_IB_IN IB_PositionData) -> return ()
      (Msg_IB_IN (IB_ManagedAccts as)) -> do
        cur <- readTVarIO $ Types.accounts currentAppData
        actions <- forM as $ \a -> case cur !? a of
                                                Just _ -> return $ id
                                                Nothing -> do
                                                  debugMsg $ "added Account "<> a
                                                  return $ HM.insertWith const a (mkIBAccount a)
        atomically $ modifyTVar (Types.accounts currentAppData) $ foldl' (.) id actions
      (Msg_IB_IN (IB_NextValidID i)) -> atomically $ modifyTVar' (nextValidID currentAppData) (const (Just i))
      (Msg_IB_IN (IB_ErrorMsg i c m)) -> debugMsg "IB_ErrorMsg not implemented"
      (Msg_IB_IN (IB_AccountValue k v c n)) -> do
        let action = HM.update (\ai -> Just $ ai & accountInfo . accountProperties %~ HM.alter (\old -> Just $ (v,c):filter ((/=c) . snd) (fromMaybe [] old)) k) n
        atomically $ modifyTVar' (Types.accounts currentAppData) action
      (Msg_IB_IN (IB_AccountUpdateTime t)) -> debugMsg "IB_AccountUpdateTime not implemented"
      -- (Msg_IB_IN (IB_AccountUpdateTime t)) -> do
      --   let action = HM.update (\ai -> Just $ ai & accountInfo . accountLastUpdate %~ const t) n
      --   atomically $ modifyTVar' (Types.accounts currentAppData) action
      (Msg_IB_IN (IB_PortfolioValue c p mp mv ac u r n)) -> do
        let cid = conId :: IBContract -> Int
            updateAction (a@IBPortfolioValue{..}:as)
             | cid _contract == cid c = IBPortfolioValue c p mp mv ac u r:as
             | otherwise              = a:updateAction as
            updateAction [] = [IBPortfolioValue c p mp mv ac u r]
            action = HM.update (\ai -> Just $ ai & accountPortfolio %~ updateAction) n
        atomically $ modifyTVar' (Types.accounts currentAppData) action
      (Msg_IB_IN (IB_SymbolSamples r s)) -> do
        atomically $ do
          modifyTVar' (nextValidID currentAppData) (const $ Just r)
          modifyTVar' (symbolLookupResults currentAppData) (const $ (\IB_SymbolSample{..} -> IBSymbolSample symId symbol secType primaryExchange currency derivatives) <$> s)
      (Msg_IB_IN t@IB_TickPrice{}) -> run $ handleTickPrice t
      _ -> --D.trace ("not implemented in AppFiller:" <> show input) $
           infoMsg $ "not implemented in AppFiller:" <> T.pack (show input)


handleTickPrice :: IB_IN -> RIO App ()
handleTickPrice IB_TickPrice{..} = do
  charts <- appCharts . appRefs <$> ask
  tid2symbol <- tickerIdToSymbol . appRefs <$> ask
  msymbol <- (HM.!? tickerId) <$> liftIO (readTVarIO tid2symbol)
  case msymbol of
    Nothing -> return () --ignore
    Just s  -> do
        chartVar <- (HM.!s) <$> liftIO (readTVarIO charts)
        case tickType of
          IBTickType_Last_Price -> do
              t <- utctDayTime <$> liftIO getCurrentTime
              let cp = ChartPoint (TimePoint $ fromInteger. (`div` 1000000000000) . diffTimeToPicoseconds $ t) price []
              liftIO $ atomically $ modifyTVar chartVar (\c@Chart{..} -> c { chartData = chartData |> cp, chartDirty = True})
          _ -> return ()
handleTickPrice _ = error "impossible"


