{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE DerivingVia #-}
module Chart (newChart, FillerException(..)) where

import Import
import Data.Time
import RIO.List
import RIO.List.Partial
import Data.FingerTree (FingerTree)
import Control.Concurrent (forkIO)
import qualified RIO.ByteString as BS
-- import Control.Exception
import qualified Data.HashMap.Strict as HM
import qualified Data.FingerTree as FT

import qualified Debug.Trace as D

data FillerException = QuitFiller
  deriving Show

instance Exception FillerException

--deriving via Integer instance Hashable Day

newChart :: IBContract -> RIO App ()
newChart contract = do
  app <- ask
  let sym = (symbol :: IBContract -> Text) contract
      hmVar = appCharts . appRefs $ app
  hm <- liftIO . readTVarIO $ hmVar
  unless (sym `HM.member` hm) $ do
    c <- liftIO $ newTVarIO $ Chart FT.empty mempty undefined defChartSettings [] Nothing False
    tid <- liftIO $ forkIO $ fillChart app contract c
    liftIO $ atomically $ do
      modifyTVar' c (\Chart{..} -> let fillerThread = tid in Chart{..})
      modifyTVar' hmVar (HM.insert sym c)

fillChart :: App -> IBContract -> TVar Chart -> IO ()
fillChart app contract cVar = runRIO app $ do
  let sym = (symbol :: IBContract -> Text) contract
  (tickerMapVar :: TVar (HashMap Int Text)) <- tickerIdToSymbol. appRefs <$> ask
  alreadyAdded <- (sym `elem`) . HM.elems <$> liftIO (readTVarIO tickerMapVar)
  unless alreadyAdded $ do
    tickerId <- (+1) . foldl' max 1000 . HM.keys <$> liftIO (readTVarIO tickerMapVar)
    let cancelSubscription = liftIO $ atomically $ do
         modifyTVar tickerMapVar (HM.delete tickerId)
         -- TODO: send cancel-request
    let sendQ = twsConnectionSend $ twsConnectionRefs $ appRefs $ app
    liftIO $ atomically $ do
      modifyTVar tickerMapVar (HM.insert tickerId sym)
      writeTQueue sendQ $ Msg_IB_OUT $ IB_RequestMktData tickerId contract "233" False False
    handle (\QuitFiller -> cancelSubscription >> exitSuccess) $
      forever $ do
        -- chart dirty? set clean & begin work
        Chart{..} <- liftIO (readTVarIO cVar)
        when chartDirty $ do
          liftIO $ atomically $ modifyTVar cVar (\c -> c { chartDirty = False })
          let (TimePoint cacheUpdateStart) = fromMaybe (TimePoint 0) Nothing -- TODO: lastCacheUpdate
              cacheUpdateEnd   = 86400
              chunkChart :: Int -> Int -> Int -> FingerTree TimePoint ChartPoint -> [(TimePoint,[ChartPoint])]
              chunkChart from until range tree = go from range interval
                  where
                    lastItem = case FT.viewr interval of
                                FT.EmptyR -> until
                                (_ FT.:> ChartPoint{..}) -> (\(TimePoint x) -> x) timeOfDay
                    interval = FT.takeUntil (\(TimePoint x) -> x > until)
                             . FT.dropUntil (\(TimePoint x) -> x > from)
                             $ tree
                    go f i t
                      | f+i >= lastItem = [(TimePoint (f+i), toList t)]
                      | otherwise       = let (a, b) = FT.split (\(TimePoint x) -> x > f+i) t
                                          in (TimePoint (f+i),toList a) : go (f+i) i b
              chunkedChart = chunkChart cacheUpdateStart cacheUpdateEnd (chartResolution chartSettings) chartData
              cachePoints = takeWhile (\ChartPoint{..} -> (\(TimePoint x) -> x < cacheUpdateStart) timeOfDay) chartCache <> map toCachePoint chunkedChart
              toCachePoint :: (TimePoint,[ChartPoint]) -> ChartPoint
              toCachePoint (t,[]) = ChartPoint t (-1) []
              toCachePoint (t,as) = ChartPoint t c [OLHC o l h c]
                where
                  as' = pointValue <$> as
                  o = head as'
                  c = last as'
                  l = minimum as'
                  h = maximum as'
          let lUpdate = fmap fst . lastMaybe $ chunkedChart
          liftIO $ atomically $ modifyTVar cVar (\c -> c { chartCache = cachePoints, lastCacheUpdate = lUpdate })
          return ()
        threadDelay 1000000 -- sleep 5 seconds
